#!/usr/local/bin/perl5

# Package drspeech_args: argument parsing for the 90's
#
# $Header: /u/drspeech/src/babylon/RCS/drspeech_args.pm,v 1.16 1998/06/18 18:34:32 gildea Exp $
#
# $Log: drspeech_args.pm,v $
# Revision 1.16  1998/06/18 18:34:32  gildea
# Minor syntax error: {} instead of () for empty hash.
#
# Revision 1.15  1998/06/02 22:40:52  fosler
# bug in substitute strings
#
# Revision 1.14  1998/06/02 22:07:18  fosler
# Changed the default behavior to write variables into the calling
# package's namespace, rather than package main.  Shouldn't affect any
# of the scripts currently written, but will allow for easier argument
# parsing in libraries.
#
# Revision 1.13  1996/11/07 21:45:03  fosler
# *** empty log message ***
#
# Revision 1.12  1996/11/07 21:42:49  fosler
# *** empty log message ***
#
# Revision 1.11  1996/11/07 21:33:09  fosler
# *** empty log message ***
#
# Revision 1.10  1996/10/30 23:05:09  fosler
# fixed error message
#
# Revision 1.9  1996/09/20 20:25:13  bedk
# Tried to clarify the description of the params variable.
#
# Revision 1.8  1996/09/13 17:42:58  fosler
# Revised so that there are no compiler warnings (with -w flag).
#
# Revision 1.7  1996/09/13 04:50:37  davidj
# Tidied up some error messages.
#
# Revision 1.6  1996/09/13 04:27:08  davidj
# Changed error messages.
#
# Revision 1.5  1996/09/13 04:21:59  davidj
# Fixed error messages.
#
# Revision 1.4  1996/09/13 00:19:07  davidj
# Updated some of the arg. checking patterns.
# Improved formatting of the usage message.
# Replaced patter_name array with pattern2name function.
# Tidied up some error messages.
#
# Revision 1.3  1996/09/06 05:40:47  fosler
# added logfile, reporting arguments
# replaced $0 with $drspeech::progname
#
# Revision 1.2  1996/09/06 01:31:05  fosler
# added documentation
#
# Revision 1.1  1996/09/05 20:01:03  fosler
# Initial revision
#

package drspeech_args;

=head1 NAME

drspeech_args - a argument parsing package

=head1 SYNOPSIS

  use drspeech;
  use drspeech_args;

=head1 DESCRIPTION

This package provides a number of routines for parsing command lines,
variable checking, and the like.

=head1 DEFINED VARIABLES

The following variables are defined by C<drspeech_args> when the
package is used:

=over 4

=item C<%drspeech_args::pattern>

An associative array that defines a mapping between pattern names and
regular expressions which describe the pattern.  See below for a
description on how to use these patterns.  Additional patterns can be
registered by defining:

C<$drspeech_args::pattern{"name"}="regular expression";>

The following patterns are predefined:

	  any
	  boolean
	  integer
	  cardinal
	  float
	  label
          opt_label
	  filename
          opt_filename
	  integer_range
	  cardinal_range
	  float_range
	  boolean_list
	  integer_list
	  cardinal_list
	  float_list
	  label_list

=item C<%drspeech_args::&pattern2name($patt)>

A routine the performs the inverse mapping of C<%drspeech_args::pattern>,
converting between regular expressions and pattern names.  See below for a
description on how to use these patterns.  This is used to print a
concise name for particular patterns.

=back

=cut

$bool = "((1)|(0)|([tT]([rR][uU][eE])?)|([fF]([aA][lL][sS][eE])?)|([yY]([eE][sS])?)|([nN][oO]?))";
$sint = "[-]?[0-9]+";
$uint = "[0-9]+";
$label = "[_a-zA-Z][0-9_a-zA-Z]*";
$float = "([-]?[0-9]+(\\.[0-9]*(e[+-]?[0-9]+)?)?)|([-]?\\.[0-9]+(e[+-]?[0-9]+)?)";
%pattern=(
	  "any" , ".*",
	  "boolean", "$bool",
	  "integer" , "$sint",
	  "cardinal", "$uint",
	  "float" , "$float",
	  "label" , "$label",
	  "opt_label" , "(${label})|",
	  "filename", ".+",
	  "opt_filename", "(.+)|",
	  "integer_range" , "${sint}[-]${sint}",
	  "cardinal_range" , "${uint}[-]${uint}",
	  "float_range", "${float}[-]${float}",
	  "boolean_list", "(${bool},)*${bool}",
	  "integer_list" , "(${sint}([-]${sint})?,)*${sint}([-]${sint})?",
	  "cardinal_list" , "(${uint}([-]${uint})?,)*${uint}([-]${uint})?",
	  "float_list", "(${float},)*${float}",
	  "label_list", "(${label},)*${label}",
);
undef($bool);
undef($sint);
undef($uint);
undef($label);
undef($float);

# Conver a regular expression for a pattern to the corresponding pattern name

sub pattern2name {
    my $p = shift;
    local $n;
    local $res;

    $res = $p;
    foreach $n (keys %pattern)
    {
	 if ($pattern{$n} eq $p)
	 {
	     $res = $n;
	 }
    }
    $res;
}

=head1 USER-DEFINED VARIABLES

Following are some variables defined by the user:

=over 4

=item C<$args>

This is a reference to a two dimensional array which gives information
about all of the arguments to be used.  This variable must be passed
into any of the parsing functions (see below).  The variable can be
named arbitrarily, since it is passed by reference into all of the
functions.  The format of the array is:

  $args=[["variable name 1","description 1",
             "syntax expression 1"],
	 ["variable name 2","description 2",
             "syntax expression 2", "default"]];

Several variables are defined to have a special meaning when included
in this array:

  logname: the name of log file, should default to "-"
    --changes the value of $drspeech::logname
  verbose: the verbosity level, should default to 1
    --changes the value of $drspeech::verbose
  params: the name of a parameters file, no default
    --a parameters file contains arguments in the form a=b
    --when the params variable is encountered, the specified file is
    --immediately processed.

The option "help", if used on the command line, forces an immediate
call to &drspeech_args::usage and exits with a status of 0.  Defining
it in C<$args> has no effect.

=over 4

=item *

Variable names: A string in the variable name position refers to a possible command
line argument as well as the name of the perl variable which holds the
value of the argument.  All variables are placed in the calling package 
(by default, package main).

=item *

Descriptions: Each variable name is accompanied by a description of that variable,
which will be printed out in the C<&drspeech_args::usage> and
C<&drspeech_args::print_args> functions. Any diagnostics or errors given
to users should have this description printed out with the variable name.

=item *

Syntax expressions:  These are regular expressions which determine if
a value of a variable is syntactically well-formed.  One can use the
predefined expressions in C<%drspeech_args::pattern> (see above) here,
or provide a regular expression in a string.  All regular expressions
match an B<entire> value (i.e. the expression is anchored to both ends
of the string).

=item *

Defaults: if an optional default is provided, then the variable takes
on that value if it is not set during argument parsing.  If it is not
provided, then the variable must be set during argument parsing.

=back

An example of an C<$args> array follows:

  $vars=[["verbose","Verbosity level",
            $drspeech_args::pattern{"integer"},1],
         ["logname","Log file",
            $drspeech_args::pattern{"filename"},"-"],
         ["params","Parameter file",
            $drspeech_args::pattern{"filename"}],
         ["foo","Foo arg",".*"],
         ["bar","Bar arg",
            $drspeech_args::pattern{"integer_list"},1],
         ["baz","Boolean baz",
            $drspeech_args::pattern{"boolean"},"false"]
         ];

=back

=head1 FUNCTIONS

=over 4

=item C<&drspeech_args::parse($args,@argv)>

This function takes an C<$args> structure as its first argument, and
interprets the rest of its arguments as an C<@ARGV> style array, with
each argument taking up one line.  It interprets each argument
C<"foo=bar"> by setting variable C<$foo> to the value C<"bar">.

It calls the following functions, in this order:
    C<&drspeech_args::checkdb_syntax>
    C<&drspeech_args::parse_args>
    C<&drspeech_args::validate>

The return value is a reference to an associative array which tells
where each variable was defined.  

=cut

sub get_calling_package {
    local @callerreturn;
    local $counter;

    for($counter=0;((@callerreturn=caller($counter)) &&
		    ($callerreturn[0] eq "drspeech_args"));
	$counter++) {}
    if (defined(@callerreturn)) {
	return $callerreturn[0];
    } else {
	return "main";
    }
}
  

sub parse {
    my $args=shift(@_);
    my @argvector=@_;
    my %argdefined=();
    local $arg;
    local $variable;
    local @array;
    local $retval;
    local $pkge;

    $pkge=&get_calling_package();

    foreach $arg (@{$args}) {
	@array=@{$arg};
	if ($#array==3) {
	    $argdefined{$array[0]}="default";
	} else {
	    $argdefined{$array[0]}="required";
	}
    }
    
    &checkdb_syntax($args);
    $retval=&parse_args("command line",$args,\%argdefined,@argvector);
    &validate($args);
    if (defined($drspeech::logname)) {
	&drspeech::open_log();
	&drspeech_args::print_args($args,$retval,$drspeech::log) if $drspeech::verbose>0;
    }
    
    return $retval;
}

sub parse_args {
    my $filename=shift(@_);
    my $args=shift(@_);
    my $argdefined=shift(@_);
    my @argvector=@_;
    my($arg,$value,$variable);
    my $counter=1;
    my $errors=0;
    my $evalstring;
    local $pkge;

    $pkge=&get_calling_package();
   
    while($#argvector>=0) {
	$arg=shift(@argvector);
	#check to see if it is in the foo=bar format
	if (($variable,$value)=($arg=~/^([^=]+)=(.*)$/)) {
	    #first off, see if is the help variable
	    if ($variable eq "help") {
		&usage($args);
		exit(0);

            #next, see if it is in the valid variable list
	    } elsif (!defined($argdefined->{$variable})) {
		warn ("unknown argument \'$variable\'.\n");
		$errors++;

	    # next check for variables with special meanings
	    } elsif ($variable eq "params"){
		&parse_file($value,$args,$argdefined);
	    } elsif ($variable eq "verbose") {
		if (!($value =~ /^$pattern{"integer"}$/)) {
		    warn ("\'verbose\' argument must be an integer.\n");
		    $errors++;
		} else {
		    $drspeech::verbose=$value;
		    $argdefined->{$variable}=$filename;
		}
	    } elsif ($variable eq "logname") {
		$drspeech::logname=$value;
		$argdefined->{$variable}=$filename;

            # finally evaluate it like a normal variable
	    } else {
		$evalstring= '$' . "$pkge"."::$variable = " . '$' . "value; ";
		eval $evalstring;
		$argdefined->{$variable}=$filename;
	    }
	} elsif ($arg eq "") {
	    if ($filename eq "command line") {
		warn ("argument/value pair $counter is empty, ignoring.\n");
	    }
	    # otherwise skip, this is a place holder from the file
	} else {
	    if ($filename eq "command line") {
		die("syntax error on command-line, argument/value pair $counter.\n");
	    } else {
		die("syntax error in file \'$filename\', line $counter\n");
	    }
	}
	$counter++;
    }
    if ($errors) {
	die ("exiting with $errors errors.\n");
    }
    return $argdefined;
}
	
sub parse_file {
    my $filename=shift(@_);
    my $args=shift(@_);
    my $argdefined=shift(@_);
    my @argvector=();
    my($variable, $value);

    open(DRSPEECH_ARGS_PARAM_FILE,"$filename") || die("can't open param file \'$filename\' for reading");
    while (<DRSPEECH_ARGS_PARAM_FILE>) {
	s/\n$//;
	if (/^#/) {
	    push(@argvector,"");
	    next;
	} elsif (/^[ 	]*$/) {
	    push(@argvector,"");
	    next;
	} else {
	    s/^[ 	]+//;
	    s/[ 	]+$//;
	    push(@argvector,$_);
	}
    }
    close(DRSPEECH_ARGS_PARAM_FILE);
    &parse_args($filename,$args,$argdefined,@argvector);
}

# checkdb_syntax:
# Do some checks to make sure that the database is kosher
#
# Database entries should be of the following form:
# [ "varname", "Variable Description", syntax-check]
# or
# [ "varname", "Variable Description", syntax-check, default-value ]
#

sub checkdb_syntax {
    my $self=shift;
    local $counter=1;
    local @array;
    local $errors=0;
    local $ref;

    foreach $ref (@{$self}) {
	@array=@{$ref};
	if ($#array < 2 || $#array > 3)  {
	    warn("botched table entry, line $counter of arg table (must have 3 or 4 args).\n");
	    $errors++;
	}
	if ($array[2] eq "") {
	    warn("NULL syntax pattern on line $counter of arg table not allowed");
	    $errors++;
	}
	$counter++;
    }

    die("exiting with $errors errors.\n") if $errors;
}

# Validate will check for all required arguments, in caller package
# It will also check the types of the arguments

sub validate {
    my $self=shift;
    local @array;
    local $errors=0;
    local $variable;
    local $newerror;
    local $syntax_check;
    local $ref;
    local $pack;
    local $value;
    local $pkge;

    $pkge=&get_calling_package();

    foreach $ref (@{$self}) {
	@array=@{$ref};
	$variable=$array[0];
	next if ($variable eq "params");
	$pack=$pkge;
	$pack="drspeech" if ($variable eq "logname" || $variable eq "verbose");
	$newerror=0;
	eval "\$newerror=1 if (!defined(\$".$pack."::$variable));";
	if ($newerror) {
	    if (defined($array[3])) {
		eval "\$".$pack."::$variable=\"$array[3]\";";
		$newerror=0;
	    } elsif ($variable ne "params") {
		warn("option $array[0] required but undefined.\n");
		$errors++;
	    }
	}
	if (! $newerror) {
	    eval "\$syntax_check=(\$".$pack."::$variable =~ /^$array[2]\$/);";
	    if (! $syntax_check) {
		eval "\$value = \$".$pack."::$variable;";
		warn "${variable}=${value}, must be of type \'" . &pattern2name($array[2]) . "\'.\n";
		$errors++;
	    }
	}
    }

    die("exiting with error count of $errors.\n") if $errors;
}

=item C<&drspeech_args::usage($args)>

=item C<&drspeech_args::usage($args,$filepointer)>

This function takes an <$args> structure and an optional reference to
a file pointer, and prints a usage statement detailing all of the
arguments to that filepointer.  Default filepointer is C<STDERR>.

=cut
    
sub usage {
    my $self=shift;
    my $filepointer=$_[0] || \*STDERR;
    local $ref;
    local $variable;
    local $type;
    local $default;  #used to get around warning messages

    print $filepointer "Usage: ${drspeech::progname} [<options>] \n";
    print $filepointer "where <options> is one or more of:\n";
    foreach $ref (@{$self}) {
	@array=@{$ref};
	$type = pattern2name($array[2]);

	if (defined($array[3])) {
	    $default="\"$array[3]\"";
	} elsif ($array[0] eq "params") {
	    $default="*OPTIONAL*";
	} else {
	    $default="*REQUIRED*";
	}

	printf $filepointer "%-40s %s",
	    "  ${array[0]}=<${type}> (${default})", "${array[1]}\n";

    }
}

=item C<&drspeech_args::print_args($args,$argdefined)>

=item C<&drspeech_args::print_args($args,$argdefined,$filepointer)>

This function prints the command-line variables to C<$filepointer>,
including information on the current value and where the variable was set
during parsing.  Default is to print to C<$drspeech::log>, or, if that
does not exist, C<STDOUT>.

=cut

sub print_args {
    my $self=shift;
    my $argdefined=shift;
    my $filepointer=$_[0] || $drspeech::log || \*STDOUT;
    local $ref;
    local $variable;
    local $evalstring;
    local $value;
    my $pack;
    local $pkge;

    $pkge=&get_calling_package();

    printf ($filepointer "$drspeech::progname: Program Variables:\n\n");
    printf ($filepointer "%-20s %-20s %-30s\n","Variable","Where assigned","Value");
    foreach $ref (@{$self}) {
	@array=@{$ref};
	$variable=$array[0];
	next if ($variable eq "params");
	$pack=$pkge;
	$pack="drspeech" if ($variable eq "verbose" || $variable eq "logname");
	$evalstring = '$'. 'value = $'.$pack."::".$variable.';';
	eval $evalstring;
	printf ($filepointer "%-20s %-20s %-30s\n",$variable,$argdefined->{$variable},$value);
    }
}

=item C<&drspeech_args::convert_list_to_array($string)>

This function will take an integer range list, and convert it into an
array of integers.  No type checking is done, so you should make sure
that what you hand this function is an integer range list.  Of course,
there are regular expressions that do this, hint, hint.

=cut

sub convert_list_to_array {
    my $string=shift;
    my $sorted = $_[0];
    my @parts;
    my @final=();
    my $part;

    @parts=split(/,/,$string);
    foreach $part (@parts) {
	$part=~s/-/../;
	eval 'push(@final,'.$part.');';
    }
    return @final;
}

=item C<&drspeech_args::convert_string_to_bool($string)>

This function takes an arbitrary string, tries to interpret it as a
boolean value, and returns 1 or 0 based on the result.  T, TRUE, Y,
YES, and positive integers map to 1.  F, FALSE, N, NO, and 0 map to
0.  The function is case-insensitive.

=cut

sub convert_string_to_bool {
    local($_) = $_[0];          # put first arg into default variable
    return 1 if /^[TtYy]/i;
    return 1 if /^[-]?[0-9]+$/ && ($_>0);
    return 0;
}

=item C<&drspeech_args::substitute_strings(@array)>

This function takes a set of strings, and replaces escape sequences of
the form C<"%n">.  It assumes the existence of the associative array
C<%string_substitution_table>, which gives a mapping from single
character escape codes to variable names.  It is the responsibility of
the programmer to maintain this table.  If a substitution is attempted
when there is no mapping, the substitution function dies.  The
sequence C<"%%"> is mapped to a single C<"%">.  This function returns
an array of mapped strings in an array context; it returns the
first mapped string in a scalar context.

=cut

sub substitute_strings {
    my @strings=@_;
    my $counter;
    my $i;
    my($value, $variable);
    my $errors=0;
    my $savestring;
    local $pkge;

    $pkge=&get_calling_package();
    
    for($i=0;$i<=$#strings;$i++) {
	$counter=0;
	$savestring=$strings[$i];
	while(($counter=index($strings[$i],"%",$counter))>=0) {
	    $value=substr($strings[$i],$counter+1,1);
	    if ($value eq "%") {
		substr($strings[$i],$counter,2)="%";
		$counter++;
	    } elsif (!defined($string_substitution_table{$value})) {
		warn("unknown sequence \%$value in string \'$savestring\'.\n");
		$errors++;
		$counter+=2;
	    } else {
		eval "\$variable = \$".$pkge.'::' . $string_substitution_table{$value} . ';';
		substr($strings[$i],$counter,2)=$variable;
		$counter+=length($variable);
	    }
	}
    }
    die("exiting with error count $errors.\n") if ($errors);
    if (wantarray) {
	return @strings;
    } else {
	return $strings[0];
    }
}

=head1 AUTHOR

Eric Fosler, fosler@icsi.berkeley.edu

=head1 SEE ALSO

drspeech(1)

=cut

1;
