#!/usr/local/bin/perl5
 
#
# noway_dict.pm: Reads in noway dictionaries
#
# Written by J. Eric Fosler-Lussier (fosler@icsi.berkeley.edu)
#
# Dictionary is in format:
#    $dict->{"word"}=[[$prior1,$ph1.1,$ph1.2,...],[$prior2,$ph2.1,...],...];
#
# $Header: /u/drspeech/src/lexicon/grapheme2phoneme/gen_efsg/RCS/efsg_dtree.pm,v 1.11 2000/02/16 23:15:55 fosler Exp $

package efsg_dtree;
use drspeech;
use onward_fsg;

$eval_debug=0;

# new(): generic new operator for blessing
sub new {
    my $this = shift;
    my $class = ref($this) || $this;
    my $self = {};
    bless $self, $class;
    return $self;
}

sub read_filelist {
    my $self = shift;
    my $treelist = shift;
    my $prunelevel = shift;
    my $nullpenalty = shift;
    my @trees;
    my $tree;
    my $base;
    my $attr;
    my $topnode;
    my $curnode;
    my $curdepth;
    my $dist;
    my @x;

    $nullpenalty=1 if !defined($nullpenalty);

    #get the list of tree files
    open(TREELIST,$treelist) || die("$0: open $treelist: $!");
    @trees=<TREELIST>;
    close(TREELIST);
    chop(@trees);

    #clear out the array
    $self->{"classes"}={};
    $self->{"topnode"}={};
    $self->{"distcounter"}={};
    $self->{"attrcounter"}=[];
    $self->{"dist"}={};

    #read in each tree
    foreach $tree (@trees) {
	$base=$tree;
	$base=~s/^.*\///;
	$base=~s/\.treeascii$//;
	if ($drspeech::verbose>=2) {
	    print STDERR "$base..";
	}

	# read in the attr file first to get the classes
	$attr=$tree;
	$attr=~s/\.treeascii$/.treeattr/;
	open(ATTR,$attr) || die("Can't open $attr for reading: $!");
	$contexts=0;
	$self->{"attr2num"}={};
	$self->{"num2attr"}=[];
	while(<ATTR>) {
	    chop;
	    if (/^class *:/) {
		s/^class *: *UNK\./class : NULL./;
		s/^class *: *//;
		s/[,\.]/ /g;
		@cl=(grep(!/UNK/,split));
		$self->{"classes"}->{$base}=[@cl];
	    } elsif (/^contexts *:/) {
		$contexts=1;
	    } elsif (!$contexts) {
		($attrn,$rest)=split(/ *:/);
		if (!defined($self->{"attr2num"}->{$attrn})) {
		    push(@{$self->{"num2attr"}},$attrn);
		    $self->{"attr2num"}->{$attrn}=$#{$self->{"num2attr"}};
		}
	    }
	}
	close(ATTR);

	# now read in the tree
	open(TREE,"$tree") || die("$0: opening tree $tree: $!");
	$topnode=[0,"0","$base",undef];
	$curnode=$topnode;
	$curdepth=-1;

	while(<TREE>) {
	    chop;
	    next if /^$/;
	    # check for distributions with only one leaf
	    if (!/:/ && $curnode==$topnode) {
		s/[(]from parent[)]//;
		$dist=$self->recorddist($base,$prunelevel,$nullpenalty,split);
		$topnode->[3]=$dist;
		next;
	    }
	    # remove colons following class list
	    s/,\010 *: */ / || s/ *: */ /;
	    # split into compontent parts
	    @x=split;
	    # count the depth (number of | in front), and remove depth markers
	    $depth=&countdepth(@x);
	    splice(@x,0,$depth);

	    # is this ever reached?
	    if ($depth==$curdepth && $curdepth==-1) {
		$topnode->[3]=join(" ",@x);
		next;
	    }

	    # check for determination between class and real attributes
	    if ($x[1] eq "in") {
		$symbols=&array2assoc(split(/,/,$x[2]));
	    } else {
		$symbols=$x[2];
	    }

	    # change attribute name into attribute number
	    $x[0]=$self->{"attr2num"}->{$x[0]};

	    # add one to the attribute counter
	    $self->{"attrcounter"}->[$x[0]]++;

	    # have we gone further down the tree?
	    if ($depth>$curdepth) {
		# yes, now check for a distribution
		if ($#x>2) {
		    #found one, record it.
		    $dist=$self->recorddist($base,$prunelevel,$nullpenalty,@x[3..$#x]);
		    #save parent, attr num, split syms, and distribution
		    $newnode=[$curnode,$x[0],$symbols,$dist];
		} else {
		    #save parent, attr num, split syms, no distribution
		    $newnode=[$curnode,$x[0],$symbols,undef];
		}
		# save as left node of curnode (< for real attributes)
		$curnode->[4]=$newnode;
		$curdepth=$depth;
		$curnode=$newnode;
	    } else {
		# how far back up the tree did we go?
		# must be the right hand side of something
		$parent=$curnode->[0];
		for($i=0;$i<($curdepth-$depth);$i++) {
		    $parent=$parent->[0];
		}
		# check for a distribution
		if ($#x>2) {
		    # found one, now record it
		    $dist=$self->recorddist($base,$prunelevel,$nullpenalty,@x[3..$#x]);
		    #save parent, attr num, split syms, and distribution
		    $newnode=[$parent,$x[0],$symbols,$dist];
		} else {
		    #save parent, attr num, split syms, no distribution
		    $newnode=[$parent,$x[0],$symbols,undef];
		}
		# save as left node of parent (>= for real attributes)
		$parent->[5]=$newnode;
		$curdepth=$depth;
		$curnode=$newnode;
	    }
	}
	close(TREE);
	$self->{"topnode"}->{$base}=$topnode;
	#if we're being real verbose, prove that we read in the tree
	if ($drspeech::verbose >= 4) {
	    print STDERR "TREE FOR $base:\n";
	    &printtree(0,$self->{"topnode"}->{$base});
	}

    }	    
    print STDERR "\n" if $drspeech::verbose>=2;
}

sub attrused {
    my $self=shift;
    my $attr=shift;

    if (defined($self->{"attr2num"}->{$attr})) {
	return $self->{"attrcounter"}->[$self->{"attr2num"}->{$attr}];
    } else {
	return 0;
    }
}

sub recorddist {
    my $self=shift;
    my $base=shift(@_);
    my $prunelevel=shift(@_);
    my $nullpenalty=shift(@_);
    my @probs=@_;
    my @dist;
    my $d;
    my $distname;
    my $i;
    my $p;
    my $deletebase=$self->{"deletebase"};
    
    $deletebase=0 if !defined($deletebase);

    #find a new name for this distribution
    $d=$self->{"distcounter"}->{$base}+0;
    $self->{"distcounter"}->{$base}++;
    $distname=$base."_".$d;

    #find out if there is an UNKnown
    local @classes=@{$self->{"classes"}->{$base}};
    if ($#classes!=$#probs) { # then we have an UNK, last prob;
	pop(@probs);
	@probs=&renormalize(@probs);
    }

    #penalize NULLs now and remove if necessary
    for($i=0;$i<=$#probs;) {
	if ($classes[$i] eq "NULL") {
	    $probs[$i]*=$nullpenalty;
	    if ($probs[$i]<$prunelevel || $probs[$i]==0) {
		splice(@probs,$i,1);
		@probs=&renormalize(@probs);
		splice(@classes,$i,1);
		next;
	    }
	}
	$i++;
    }

    #check to see if any probs are under prunelevel
    for($i=0;$i<=$#probs;) {
	if (($probs[$i]<$prunelevel || $probs[$i]==0) && 
	    $#probs>0 &&
	    ($deletebase || $classes[$i] ne $base)) {
	    splice(@probs,$i,1);
	    splice(@classes,$i,1);
	} else {
	    $i++;
	}
    }
    @probs=&renormalize(@probs);

    for($i=0;$i<=$#probs;$i++) {
	$dist[$i]=[$probs[$i],split(/_/,$classes[$i])];
    }
    $self->{"dist"}->{$distname}=[@dist];
    return $distname;
}

sub printtree {
    local $depth=shift(@_);
    return if !defined($_[0]);
    local $tree=shift(@_);
    local $fh=shift(@_);

    $fh=\*STDERR if !defined($fh);

    $syms=join(",",keys(%{$tree->[2]}));
    local $i;
    for($i=0;$i<$depth;$i++) {
	print $fh "  ";
    }
    if (defined($tree->[3])) {
	print $fh $tree->[1]," ",$syms," ",$tree->[3],"\n";
    } else {
	print $fh $tree->[1]," ",$syms,"\n";
    }
    &printtree($depth+1,$tree->[4]);
    &printtree($depth+1,$tree->[5]);
}

sub evaluate {
    my $self=shift;
    my $center=shift;
    my @letters=@_;
    
    &drspeech::inform("Evaluating $center") if ($eval_debug);
    return &evaluate_iter($self->{"topnode"}->{$center},@letters);
}

sub treedefined {
    my $self=shift;
    my $treename=shift;

    return defined($self->{"topnode"}->{$treename});
}

sub attr2num {
    my $self=shift;
    my $attrname=shift;
    
    return $self->{"attr2num"}->{$attrname};
}

sub evaluate_iter {
    local $tree=shift(@_);
    local @letters=@_;

    #recurse down the tree, using the splitting criteria in the tree

    #if we have a distribution, we're at a leaf... stop.
    if (defined($tree->[3])) {
	&drspeech::inform("Returning leaf ".$tree->[3]) if ($eval_debug);
	return $tree->[3];
    }
    # number of attribute is in $l
    local $l=$tree->[4]->[1];
    # split criteria: associative arrays if class, numbers if real attr
    local $leftside=$tree->[4]->[2];
    local $rightside=$tree->[5]->[2];

    #check for type of criterion
    if (ref($leftside) eq "HASH") {
	# found a class hash
	&drspeech::inform("Evaluating string attr ".$l."=".$letters[$l]) if ($eval_debug);
	if (defined($leftside->{$letters[$l]})) {        #go left?
	    &drspeech::inform("  ..left branch") if ($eval_debug);
	    return &evaluate_iter($tree->[4],@letters);
	} elsif (defined($rightside->{$letters[$l]})) {  #or right?
	    &drspeech::inform("  ..right branch") if ($eval_debug);
	    return &evaluate_iter($tree->[5],@letters);
	} elsif (defined($leftside->{"UNK"})) {          #UNK on left?
	    &drspeech::inform("  ..left branch (UNK)") if ($eval_debug);
	    return &evaluate_iter($tree->[4],@letters);
	} elsif (defined($rightside->{"UNK"})) {         #UNK on right?
	    &drspeech::inform("  ..right branch (UNK)") if ($eval_debug);
	    return &evaluate_iter($tree->[5],@letters);
	} else {                                         #give up.
	    die("evaluate: unknown symbol for feature #$l: $letters[$l] (expected one of ".join(",",keys %{$leftside},keys %{$rightside}).")");
	}
    } elsif (!ref($leftside)) {
	&drspeech::inform("Evaluating real attr ".$l."=".$letters[$l]) if ($eval_debug);
	# found a real attribute
	if ($letters[$l]<$leftside) {                    #go left if val<cut
	    &drspeech::inform("  ..left branch ($letters[$l]<$leftside)") if ($eval_debug);
	    return &evaluate_iter($tree->[4],@letters);
	} else {                                         #go right
	    &drspeech::inform("  ..right branch ($letters[$l]>=$leftside)") if ($eval_debug);
	    return &evaluate_iter($tree->[5],@letters);
	}
    } else {
	die("evaluate: unknown type found in internal tree representation.  Shouldn't get here.");
    }
}
sub evaluate_fragment {
    my $self=shift;
    local $i=shift;
    local $beginfragment=shift;
    local $trialwords=shift;
    local $nletters=shift;

    local $word;
    local $position=$nletters-$i;
    local $wdposition;
    local $e;
    local @dist;
    local @a;
    local $j;
    local $probsum=0;
    my %probs=();

    for($word=0;$word<=$#{$trialwords};$word++) {
      @wdletters=@{$trialwords->[$word]};
      if ($beginfragment) {
          $e=$self->evaluate($wdletters[$i],@wdletters[$i-3..$i+3]);
      } else {
          $wdposition=$#wdletters-$position;
          $e=$self->evaluate($wdletters[$i],@wdletters[$wdposition-3..$wdposition+3]);
      }
      @dist=@{$self->{"dist"}->{$e}};
      for($j=0;$j<=$#dist;$j++) {
          @a=@{$dist[$j]};
          $phonestring=join("_",@a[1..$#a]);
          if (!defined($probs{$phonestring})) {
              $probs{$phonestring}=[@a];
          } else {
              $probs{$phonestring}->[0]+=$a[0];
          }
      }
      $probsum++;
    }
    @dist=();
    foreach $e (values %probs) {
      $e->[0]=$e->[0]/$probsum;
      push(@dist,$e);
    }
    return @dist;
}

sub set_dictionary {
    my $self=shift;
    my $dictionary=shift;
    if ($drspeech::verbose>=3) {
	print STDERR "Setting dictionary in trees.\n";
    }
    $self->{"dictionary"}=$dictionary;
    $self->{"dictwords"}=[keys %{$dictionary}];
}

sub find_fragments {
    my $self=shift;
    my $w=shift;
    my $dictwords=$self->{"dictwords"};
    my $fragment_samplect=(defined($_[0]))?shift:10;
    my @tr=undef;
    my @trialwords=();
    my @trialwordstrings=();
    my $oldw;
    my $i;
    my $x;
    my $ww;
    my $beginfragment;

    $oldw=$w;
    if ($w=~/^[^-]+-$/) {
	$w=~s/-//;
	@tr=grep(/^$w/,@{$dictwords});
	$beginfragment=1;
    } elsif ($w=~/^-[^-]+$/) {
	$w=~s/-//;
	@tr=grep(/$w$/,@{$dictwords});
	$beginfragment=0;
    }
    if (defined(@tr)) {
	for($i=0;$#tr>=0 && $i<$fragment_samplect;$i++) {
	    $x=int(rand($#tr+1));
	    $ww=splice(@tr,$x,1);
	    push(@trialwordstrings,$ww);
	    push(@trialwords,[&word2letters($ww)]);
	}
	if ($drspeech::verbose>=3) {
	    printf STDERR ("Using fragment generator for %s with the folliwing words:\n%s\n",$oldw,join(" ",@trialwordstrings));
	}
	
    } 
    if ($#trialwords>=0) {
	return ($beginfragment,[@trialwords]);
    } else {
	return ($beginfragment,undef);
    }
}

sub buildfsg {
    use onward_fsg;
    my $self=shift;
    my $w=shift;
    my $curstate=0;
    my $dictwords=$self->{"dictwords"};
    my $fragment_samplect=$self->{"fragment_samplect"};

    my $trialwords;
    my $beginfragment;
    my @dist;
    my $maxout;
    my $e;
    my @a;
    my @letters;
    my $fsg=new onward_fsg();

    #look for word fragments
    undef $trialwords;
    if (defined($dictwords) && $w=~/-/) {
	($beginfragment,$trialwords)=
	    $self->find_fragments($w,$fragment_samplect);
    }
    $w=~s/-//;

    @letters=&word2letters($w);
    for($i=3;$i<=$#letters-3;$i++) {
	print STDERR "$letters[$i]: " if $drspeech::verbose>=2;
		
	if (!defined($trialwords)) { #just feed the original word in
	    $e=$self->evaluate($letters[$i],@letters[$i-3..$i+3]);
	    print STDERR $e,":" if $drspeech::verbose>=2;
	    @dist=@{$self->{"dist"}->{$e}};
	} else {
	    @dist=$self->evaluate_fragment($i,$beginfragment,$trialwords,$#letters);
	}
	$maxout=0;
	for($j=0;$j<=$#dist;$j++) {
	    if ($drspeech::verbose>=2) {
		@a=@{$dist[$j]};
		printf STDERR ("%s/%.6f ",join("_",@a[1..$#a]),$a[0]) if $drspeech::verbose>=2;
	    }
	    $maxout=$#{$dist[$j]} if $#{$dist[$j]}>$maxout;
	}
	print STDERR "\n" if $drspeech::verbose>=2;
	$nextstate=$curstate+$maxout;
	for($j=0;$j<=$#dist;$j++) {
	    @out=@{$dist[$j]};
	    $fsg->addtrans($curstate, 
			   ($#out==1)?$nextstate:$curstate+1,
			   $out[0],$out[1]);
	    if ($#out>1) {
		$xstate=$curstate+1;
		for($k=2;$k<=$#out;$k++) {
		    $fsg->addtrans($xstate,$xstate+1,
				   1.0,$out[$k]);
		    $xstate++;
		}
	    }
	}
	$curstate=$nextstate;
    }	
#    return $fsg->rmepsilon("NULL");
    return $fsg;
#   $fsgs[$wd]=$fsgs[$wd]->prune_unreachable();
}   

sub buildnowayentries {
    my $self=shift;
    my $w=shift;
    my $maxprons=shift;
    my $dictwords=$self->{"dictwords"};
    my $fragment_samplect=$self->{"fragment_samplect"};
    my @oldpronstack=([1.0,""]);
    my @newpronstack=();
    my $trialwords;
    my $beginfragment;
    my @dist;
    my $maxout;
    my $e;
    my @a;
    my @latters;
    my $fsg=new onward_fsg();

    $maxprons=10 if !defined($maxprons);

    #look for word fragments
    undef $trialwords;
    if (defined($dictwords) && $w=~/-/) {
	($beginfragment,$trialwords)=
	    $self->find_fragments($w,$fragment_samplect);
    }
    $w=~s/-//;

    @letters=&word2letters($w);
    for($i=3;$i<=$#letters-3;$i++) {
	print "$letters[$i]: " if $drspeech::verbose>=2;
		
	if (!defined($trialwords)) { #just feed the original word in
	    $e=$self->evaluate($letters[$i],@letters[$i-3..$i+3]);
	    print $e,":" if $drspeech::verbose>=2;
	    @dist=@{$self->{"dist"}->{$e}};
	} else {
	    @dist=$self->evaluate_fragment($i,$beginfragment,$trialwords,$#letters);
	}
	$maxout=0;
	for($j=0;$j<=$#dist;$j++) {
	    @a=@{$dist[$j]};
	    if ($drspeech::verbose>=2) {
		printf("%s/%.6f ",join("_",@a[1..$#a]),$a[0]) if $drspeech::verbose>=2;
	    }
#	    if ($a[1] eq "NULL") {
#		@a=($a[0]);
#	    }
	    foreach $oldpron (@oldpronstack) {
		push(@newpronstack,[$oldpron->[0]*$a[0],join(" ",$oldpron->[1],@a[1..$#a])]);
	    }
	}
	print "\n" if $drspeech::verbose>=2;

	my %uniqprons=();
	my ($n,$key);
	foreach $n (@newpronstack) {
	    $key=$n->[1];
	    if (defined($uniqprons{$key})) {
		$uniqprons{$key}->[0]+=$n->[0];
	    } else {
		$uniqprons{$key}=$n;
	    }
	}
	@oldpronstack=sort { $b->[0] <=> $a->[0]; } values %uniqprons;
	splice(@oldpronstack,$maxprons);
	@newpronstack=();
    }
    return @oldpronstack;
}   

# utility routines
sub array2assoc {
    local $c={};
 
    foreach $x (@_) {
        $c->{$x}=0;
    }
    return $c;
}

sub countdepth {
    local $counter=0;
    if ($_[0] =~ /^[0-9\.]+$/) {
        return -1;
    }
 
    while($_[0] eq "|") {
        shift(@_);
        $counter++;
    }
    return $counter;
}

sub renormalize {
    local @a=@_;
    local $total=0;
    local $a;

    foreach $a (@a) {
	$total+=$a;
    }
    foreach $a (@a) {
	$a=$a/$total;
    }
    return @a;
}

sub sum {
    local @a=$_;
    local $total=0;
    local $a;

    foreach $a (@a) { $total+=$_; }
    return $total;
}

sub word2letters {
    my $word=shift;
    my @latters;
    my $l;

    @letters=split("",$word);
    foreach $l (@letters) {
	$l=$efsg_lettermap{$l} if defined $efsg_lettermap{$l};
    }
    unshift(@letters,"HASH","HASH","HASH");
    push(@letters,"HASH","HASH","HASH");
    return @letters;
}	

%efsg_lettermap=(".","DOT",
		 "'","APOST",
		 "#","HASH",
		 "(","LPAREN",
		 ")","RPAREN",
		 "-","DASH",
		 "<","LTHAN",
		 ">","GTHAN",
		 "_","UNDERSCORE",
		 "","a2",
		 "","a3",
		 "","c5",
		 "","e2",
		 "","e1",
		 "","e3",
		 "","e4",
		 "","i3",
		 "","i4",
		 "","i3",
		 "","u2",
		 "","u3",
		 "","u4");

1;
