#!/usr/local/bin/perl5

#
# fsg.pm: A generic finite state grammar package
# 
# Written by J. Eric Fosler-Lussier (fosler@icsi.berkeley.edu)
#
# States are organized as a two dimensional array;
# states[$from] gives an array of transitions
# states[$from]->[$n] gives transition n, of the form
#    [to-state probability label1 label2 ...]
#
# WARNING: not all of the functions you might want are implemented here
#          this package will probably grow in the future
#
# $Header: /u/drspeech/src/lexicon/grapheme2phoneme/gen_efsg/RCS/fsg.pm,v 1.3 1998/09/23 18:05:26 fosler Exp $

package fsg;

# function for sorting transitions by to-state
sub bytostate {
    $a->[0] <=> $b->[0];
}

# new(): generic new operator for blessing
sub new {
    my $this = shift;
    my $class = ref($this) || $this;
    my $self = {};
    bless $self, $class;
    $self->initialize();
    return $self;
}

# initialize(): initalization routines
sub initialize {
    my $self=shift;
    $self->{"states"}=[];
    $self->{"printstart"}=0;
}

# addtrans(from,to,prob,label1,...): add a single transition
sub addtrans {
    my $self=shift;
    my $from=shift;
    my $to=shift;
    my $prob=shift;
    my @labels=@_;
    my $states=$self->{"states"};

    $states->[$from]=[] if !defined($states->[$from]);
    push(@{$states->[$from]},[$to,$prob,@labels]);
}

# settrans(from,transition-array): change all of the transitions from a node
sub settrans {
    my $self=shift;
    my $from=shift;
    my $toproblabelarray=shift;
    my $states=$self->{"states"};

    $states->[$from]=$toproblabelarray;
}

# counttrans(): count the number of transitions in the entire fsg
# counttrans(from): count the number of transitions from a state
# counttrans(from,to): count the number of transitions between two states
sub counttrans {
    my $self=shift;
    my $from=shift;
    my $to=shift;
    my $counter=0;
    my $states=$self->{"states"};
    my $tran;

    if (defined($from)) {
	if (defined($to)) {
	    foreach $tran (@{$states->[$from]}) {
		$counter++ if ($tran->[0]==$to);
	    }
	    return $counter;
	} else {
	    return $#{$states->[$from]}+1;
	}
    } else {
	for($from=0;$from<=$#{$states};$from++) {
	    $counter+=$#{$states->[$from]}+1;
	}
    }
}

# printefsg_body(): prints the body of an EFSG to stdout
# printefsg_body(filehandle): prints the body of an EFSG to filehandle
# printefsg_body(arrayref): prints the body of an EFSG to an array
sub printefsg_body {
    my $self=shift;
    my $fh=shift;

    $fh=\*STDOUT if !defined($fh);

    my $states=$self->{"states"};
    my $maxstates=$#{$states};
    my $from;
    my $to;
    my $trans;
    my $printstart=$self->{"printstart"};
    my @trans;

    for($from=0;$from<=$maxstates;$from++) {
	if (ref($fh) eq "GLOB") {
	    printf $fh ("# defn of state %d\n",$from+$printstart);
	    printf $fh ("%d\t%d\n",$from+$printstart,
			$self->counttrans($from));
	} elsif (ref($fh) eq "ARRAY") {
	    push(@{$fh},
		 sprintf("# defn of state %d\n",$from+$printstart),
		 sprintf("%d\t%d\n",$from+$printstart,
			 $self->counttrans($from)));
	} else {
	    warn("fsg::printefsg_body: passed non GLOB or ARRAY reference");
	}

	@trans=sort bytostate @{$states->[$from]};
	foreach $tran (@trans) {
	    @t=@{$tran};
	    if (ref($t[3]) eq "ARRAY") { # have a lm score for a word
		$t[1]*=$t[3]->[1];
		$t[3]=$t[3]->[0];
	    }
	    if (ref($fh) eq "GLOB") {
		printf $fh ("%d\t%.4g\t%s\n",$t[0]+$printstart,
			    $t[1],join("\t",@t[2..$#t]));
	    } elsif (ref($fh) eq "ARRAY") {
		push (@{$fh},sprintf("%d\t%.4g\t%s\n",$t[0]+$printstart,
				     $t[1],join("\t",@t[2..$#t])));
	    } 

	}
    }
}

# printefsg(): prints the header and body of an EFSG to stdout
# printefsg(filehandle): prints the header and body of an EFSG to filehandle
# printefsg(filehandle,magic): prints the header and body of an EFSG to filehandle with different magic number 
sub printefsg {
    my $self=shift;
    my $fh=shift;
    my $magic=shift;
    my $states=$self->{"states"};
    my $maxstates=$#{$states};

    $fh=\*STDOUT if !defined($fh);
    $magic="EFG" if !defined($magic);

    if (ref($fh) eq "GLOB") {
	print $fh "$magic\n";
	print $fh "# Generated by $0 on ".`date`;
	print $fh "# number of states\n";
	print $fh $maxstates+1,"\n";
    } elsif (ref($fh) eq "ARRAY") {
	push(@{$fh},
	     "$magic\n",
	     "# Generated by $0 on ".`date`,
	     "# number of states\n",
	     sprintf("%d\n",$maxstates+1));
    } else {
	warn("fsg::printefsg: passed non GLOB or ARRAY reference");
    }

    $self->{"printstart"}=0;
    $self->printefsg_body($fh);
}

#make require happy	    
1;
