#!/usr/local/bin/perl5

#
# gen_efsg.pm: converts reference transcripts into phone-level efsg using
#              grapheme-to-phoneme trees or dictionary
#
# Written by J. Eric Fosler-Lussier (fosler@icsi.berkeley.edu)
#
# $Header: /u/drspeech/src/lexicon/grapheme2phoneme/gen_efsg/RCS/gen_efsg.pm,v 1.8 2000/02/16 23:05:09 fosler Exp $

package gen_efsg;

# new(): generic new operator for blessing
sub new {
    my $this = shift;
    my $class = ref($this) || $this;
    my $self = {};
    bless $self, $class;
    $self->initialize(@_);
    return $self;
}

# initialize(): initalization routines
sub initialize {
    my $self=shift;

    use drspeech;
    use drspeech_args;
    use onward_fsg;
    use efsg_dtree;
    use efsg_dict;
    use noway_dict;

    $drspeech_args::pattern{"stm-or-ref"}="(stm)|(ref)";
    
    $args=[["verbose","Verbosity level",$drspeech_args::pattern{"integer"},1],
	   ["treelist","name of file containing list of trees",
	    $drspeech_args::pattern{"opt_filename"},""],
	   ["silprob","probability of interword silence",
	    $drspeech_args::pattern{"float"},0.001],
	   ["silwrd","silence word label",
	    $drspeech_args::pattern{"any"},"<SIL>"],
	   ["silphn","silence phone label",
	    $drspeech_args::pattern{"any"},"h#"],
	   ["print_pronv","Print pronunciation version",
	    $drspeech_args::pattern{"boolean"},"false"],
	   ["fragment_samplect",
	    "Number of random words to determine fragment pron from ",
	    $drspeech_args::pattern{"integer"},10],
	   ["random_seed","seed for initializing random number generator",
	    $drspeech_args::pattern{"float"},0],
	   ["dict","dictionary for looking up in-vocabulary forms",
	    $drspeech_args::pattern{"opt_filename"},""],
	   ["prunelevel","Prune phones with less than this probability",
	    $drspeech_args::pattern{"float"},0.01],
	   ["multiwords","Should look for multiwords from the dictionary",
	    $drspeech_args::pattern{"boolean"},"false"],
	   ["alphabet","Allowable characters in the alphabet",
	    $drspeech_args::pattern{"any"},"A-Za-z\\-\\'\\.\\#"]];



    $found=&drspeech_args::parse($args,@_);

    if ($dict eq "" && $treelist eq "") {
	die("gen_efsg::initialize: must define one of dict or treelist");
    }

    srand($random_seed);

    if ($dict ne "") {
	$dictionary=new noway_dict();
	$dictionary->read_file($dict);
	$self->{"dictionary"}=$dictionary;
	$efsgdict=new efsg_dict();
	$efsgdict->set_dictionary($dictionary);
	$efsgdict->{"print_pronv"}=&drspeech_args::convert_string_to_bool($print_pronv);
	$self->{"efsgdict"}=$efsgdict;
    }

    if ($treelist ne "") {
	$trees=new efsg_dtree();
	$trees->read_filelist($treelist,$prunelevel);
	$trees->set_dictionary($dictionary) if $dict ne "";
	$trees->{"fragment_samplect"}=$fragment_samplect;
	$self->{"trees"} = $trees;
    }

    $self->{"alphabet"}=$alphabet;
    $self->{"silprob"}=$silprob;
    $self->{"silwrd"}=$silwrd;
    $self->{"silphn"}=$silphn;
    $self->{"print_pronv"}=&drspeech_args::convert_string_to_bool($print_pronv);
    $self->{"fragment_samplect"}=$fragment_samplect;
    $self->{"prunelevel"}=$prunelevel;
    $self->{"multiwords"}=&drspeech_args::convert_string_to_bool($multiwords);

    warn("ready\n") if $drspeech::verbose>=2;
}

sub get_efsg {
    my $self=shift;
    my @words=@_;
    my $dictionary=$self->{"dictionary"};
    my $trees=$self->{"trees"};
    my $efsgdict=$self->{"efsgdict"};
    my $alphabet=$self->{"alphabet"};
    my $silprob=$self->{"silprob"};
    my $silwrd=$self->{"silwrd"};
    my $silphn=$self->{"silphn"};
    my $print_pronv=$self->{"print_pronv"};
    my $fragment_samplect=$self->{"fragment_samplect"};
    my $prunelevel=$self->{"prunelevel"};
    my $multiwords=$self->{"multiwords"};
    my $wd;

#    print STDERR "multiwords: $multiwords\n";
    my @fsgs=();
    my @initstates=(0);
    my $joined_fsg=new onward_fsg;
    my $multiword_fsg;

    for($wd=0;$wd<=$#words;$wd++) {
# choose between dictionary and tts trees
	if (defined($dictionary) && 
	    defined($dictionary->{$words[$wd]})) {    # use the dictionary
	    &drspeech::inform("Using dictionary for \"$words[$wd]\"\n") if ($drspeech::verbose>4);
	    $fsgs[$wd]=$efsgdict->buildfsg($words[$wd]);
	} elsif (defined($trees)) { # use the tts
#get rid of characters for which we don't have trees
#XXX: do these characters have effects on other characters?
	    $w=$words[$wd];
	    $w=~s/[^$alphabet]//g;
	    $w=uc($w);
	    &drspeech::inform("Building \"$w\" from trees\n") if ($drspeech::verbose>3);
	    $fsgs[$wd]=$trees->buildfsg($w);
	} else {
	    warn("gen_efsg::get_efsg: no tts trees defined and word $words[$wd] not in dictionary");
	}
	$fsgs[$wd]->setlabels("+".$words[$wd],3,0);
	if ($silprob>0 && $fsgs[$wd]->{"finalstate"}>0) {
	    $fsgs[$wd]->scaleprobs(1-$silprob,0);
	    $fsgs[$wd]->addtrans(0,0,$silprob,$silphn,"+".$silwrd);
	}
	$joined_fsg=$joined_fsg->join($fsgs[$wd]);
#	$joined_fsg->printefsg(\*STDERR);
	#check for multiwords
	if ($multiwords && $wd>0 && defined($dictionary) &&
	    defined($dictionary->{$words[$wd-1]."_".$words[$wd]})) {
#	    print STDERR "found ".$words[$wd-1]."_".$words[$wd]."\n";
	    $multiword_fsg=$efsgdict->buildfsg($words[$wd-1]."_".$words[$wd]);
	    $multiword_fsg->setlabels("+".$words[$wd-1]."_".$words[$wd],3,0);
	    $joined_fsg->addparallel($multiword_fsg,$initstates[$wd-1]);
#	    $joined_fsg->printefsg(\*STDERR);
	} elsif ($multiwords && $wd>0) {
#	    print STDERR "didn't find ".$words[$wd-1]."_".$words[$wd]."\n";
	}
	$initstates[$wd+1]=$joined_fsg->{"finalstate"};
    }
    $f=new onward_fsg();
    $f->addtrans(0,0,1.0,$silphn,"+".$silwrd);
    $joined_fsg=$joined_fsg->join($f);

    @lines=();

    &onward_fsg::join_and_print_efsg(\@lines,"EFG",$joined_fsg);
    return @lines;
}

sub get_noway {
    my $self=shift;
    my $maxprons=shift;
    my @words=@_;
    my $trees=$self->{"trees"};
    my $alphabet=$self->{"alphabet"};
    my $fragment_samplect=$self->{"fragment_samplect"};
    my $prunelevel=$self->{"prunelevel"};
#    my $maxprons=$self->{"maxprons"};
    my $nowaydict=new noway_dict();
    my $wd;
    my @prons;
    
    # iterate through all of the words passed in
    for($wd=0;$wd<=$#words;$wd++) {
	my $w;
	$w=$words[$wd];
	#remove any non-legal characters
	$w=~s/[^$alphabet]//g;
	#build dictionary entries for this word using the TTS trees
	@prons=$trees->buildnowayentries($w,$maxprons);

	#now, prune prons less than prunelevel
	my ($pron,$total);
	foreach $pron (@prons) {
	    if ($pron->[0]>$prunelevel) {
		print STDERR join(" ","adding",$words[$wd],@{$pron}),"\n" if ($drspeech::verbose>=3);
		$nowaydict->add_pron($words[$wd],@{$pron});
	    }
	}
	#now renormalize to account for pruned pronunciations
	$nowaydict->renormalize($words[$wd]);
    }
    return $nowaydict;
}
	
1;


