#!/usr/local/bin/perl5
 
#
# noway_dict.pm: Reads in noway dictionaries
#
# Written by J. Eric Fosler-Lussier (fosler@icsi.berkeley.edu)
#
# Dictionary is in format:
#    $dict->{"word"}=[[$prior1,$ph1.1,$ph1.2,...],[$prior2,$ph2.1,...],...];
#
# $Header: /u/drspeech/src/lexicon/grapheme2phoneme/gen_efsg/RCS/noway_dict.pm,v 1.5 1998/10/19 17:03:57 fosler Exp $

package noway_dict;
use drspeech;

# new(): generic new operator for blessing
sub new {
    my $this = shift;
    my $class = ref($this) || $this;
    my $self = {};
    bless $self, $class;
    if (defined($_[0])) {
	$self->read_file(@_);
    }
    return $self;
}
 
sub read_file {
    my $self = shift;
    my $file = shift;
    my $remove_markers=shift;
    my $dict = {};
    my $line;
    my $prior;
    my $word;
    my $pron;
    my @pron;
    my $errors=0;
    my $count=0;
    
    if ($drspeech::verbose >= 2) {
	print STDERR "Reading dictionary $file...";
    }
    if ($drspeech::verbose == 3) {  # at verbosity 4 and above we print the words
	print STDERR "   0K";
    }

    if (!defined($remove_markers)) {
	$remove_markers=1;
    }

    open(NOWAY_DICT_READ,$file) || die("$0: can't open $file for reading: $!");
    while(<NOWAY_DICT_READ>) {
	chop;
	next if /^#/;
	$count++;
	if ($remove_markers) {
	    s/ \+ / /g;
	    s/ \+$//;
	}
	$line=$_;
	$prior=-1;
	#first, try with prior, then try without
	if ((($word,$prior,$pron)=
	     ($line=~/^([^ 	]+)\(([0-9\.]+)\)[ 	]+(.*)$/)) ||
	    (($word,$pron)=
	     ($line=~/^([^ 	]+)[ 	]+(.*)$/))) {
	    if ($drspeech::verbose == 3 && $count % 1000 == 0) {
		printf STDERR (" %3dK",int($count/1000));
	    }
	    if ($drspeech::verbose >=4) {
		print STDERR "$word..";
	    }
	    $prior=1 if $prior==-1;
	    @pron=split(/[ 	]+/,$pron);
	    if (!defined($self->{$word})) {
		$self->{$word}=[[$prior,@pron]];
	    } else {
		push(@{$self->{$word}},[$prior,@pron]);
	    }
	} else {
	    warn("dictionary read ($file): misformatted line \"$line\"");
	    $errors++;
	}
    }
    close(NOWAY_DICT_READ);
    if ($drspeech::verbose >=2) {
	print STDERR "done.\n";
    }
    
    return $errors;
}

# add a pronunciation to the dictionary with probability prob
sub add_pron {
    my $self=shift;
    my $word=shift;
    my $prior=shift;
    my @pron=@_;

    if (!defined($self->{$word})) {
	$self->{$word}=[[$prior,@pron]];
    } else {
	push(@{$self->{$word}},[$prior,@pron]);
    }
}

#renormalize probabilities for a word, or for entire dictionary if
#word is left out
sub renormalize {
    my $self=shift;
    my $word=shift;
    my @words;
    my $w;
    my $p;
    my $total;

    #if word is given, only operate on that word, otherwise renorm everything
    if (defined($word)) {
	@words=($word);
    } else {
	@words=keys %{$self};
    }

    # iterate over all words
    foreach $w (@words) {
	#sum up probabilities
	$total=0;
	foreach $p (@{$self->{$w}}) {
	    #prob is the first element of pronunciation
	    $total+=$p->[0];
	}
	#now divide by total if over zero
	if ($total>0) {
	    foreach $p (@{$self->{$w}}) {
		$p->[0]/=$total;
	    }
	}
    }
}

sub print {
    my $self=shift;
    my $fh=shift;
    my @words=sort keys %{$self};
    my $w;
    my $p;
    
    $fh=\*STDOUT if !defined($fh);

    #iterate over all words
    foreach $w (@words) {
	foreach $p (@{$self->{$w}}) {
	    @pron=@{$p};
	    printf $fh ("%s(%.4f) %s\n",$w,$pron[0],join(" ",@pron[1..$#pron]));
	}
    }
}
    
1;
