#!/usr/local/bin/perl5

#
# onward_fsg.pm: finite state grammars where transitions are monotonically 
#                non-decreasing in states
#
# Written by J. Eric Fosler-Lussier (fosler@icsi.berkeley.edu)
#
# States are organized as a two dimensional array;
# states[$from] gives an array of transitions
# states[$from]->[$n] gives transition n, of the form
#    [to-state probability label1 label2 ...]
#
# WARNING: not all of the functions you might want are implemented here
#          this package will probably grow in the future
#
# $Header: /u/drspeech/src/lexicon/grapheme2phoneme/gen_efsg/RCS/onward_fsg.pm,v 1.8 1998/10/19 17:02:09 fosler Exp $

package onward_fsg;
#inheritance from fsg
use fsg;
@ISA= qw(fsg);

# new(): generic new operator for blessing
sub new {
    my $this = shift;
    my $class = ref($this) || $this;
    my $self = {};
    bless $self, $class;
    $self->initialize();
    return $self;
}

# initialize(): initalization routines
sub initialize {
    my $self=shift;
    $self->{"states"}=[];
    $self->{"printstart"}=0;
    $self->{"finalstate"}=-1;
}

# addtrans(from,to,prob,label1,...): add a single transition, 
#                                    check for monotonicity
sub addtrans {
    my $self=shift;
    my $from=shift;
    my $to=shift;
    my $prob=shift;
    my @labels=@_;

    if ($from>$to) {
	my @c=caller;
	warn("onward_fsg::addtrans: from state $from > to state $to, not adding trans; called from $c[0]::$c[1], line $c[2]");
	return;
    }
    if ($to>$self->{"finalstate"}) {
	$self->{"finalstate"}=$to;
	undef $self->{"penultimate_states"};
    }
    
    &fsg::addtrans($self,$from,$to,$prob,@labels);
}

# addparallel(fsg): add a parallel path starting at state 0; change
#                   final state
sub addparallel {
    my $self=shift;
    my $parallelfsg=shift;
    my $initialstate=shift;
    my $finalstate=$self->{"finalstate"};
    my $pf_finalstate=$parallelfsg->{"finalstate"};
    my $states=$self->{"states"};
    my $pf_states=$parallelfsg->{"states"};
    my ($i,$j);
    my $trans;
    my @penultimate_states;
    my %new_penultimate_states;

    if (!defined($initialstate)) {
	$initialstate=0;
    }

    if (defined($self->{"penultimate_states"})) {
	@penultimate_states=@{$self->{"penultimate_states"}};
	if ($#penultimate_states<0) {
	    @penultimate_states=(0..$#{$states}-1);
	}
#	print STDERR join(" ","ps1:",@penultimate_states),"\n";
    } else {
	@penultimate_states=(0..$#{$states}-1);
#	print STDERR join(" ","ps2:",@penultimate_states),"\n";
    }

#    print STDERR ("Joining fsg2 (0..$pf_finalstate) to fsg1 ($initialstate..$finalstate).\n");
    if ($finalstate != $#{$states}) {
	die("onward_fsg::addparallel: finalstate ($finalstate) != number of states ($#{$states})");
    }
#    for($i=0;$i<$#{$states};$i++) {
#    print STDERR join(" ","ps:",@penultimate_states),"\n";
    foreach $i (@penultimate_states) {
	$trans=$states->[$i];
	for($j=0;$j<=$#{$trans};$j++) {
	    if($trans->[$j]->[0] == $finalstate) {
		$trans->[$j]->[0] += ($pf_finalstate-1);
		$new_penultimate_states{$i}=1;
	    }
	}
    }
    for($i=0;$i<=$#{$pf_states};$i++) {
	my $newfrom=($i==0)?$initialstate:$i+$finalstate-1;
	$trans=$pf_states->[$i];
	if ($#{$trans}>=0) {
	    # do an addtrans to $self appropriately
	    for($j=0;$j<=$#{$trans};$j++) {
		my @transinfo=@{$trans->[$j]};
		$newto=($transinfo[0]==0)?$initialstate:$transinfo[0]+$finalstate-1;
		$self->addtrans($newfrom,$newto,@transinfo[1..$#transinfo]);
		$new_penultimate_states{$newfrom}=1 if $transinfo[0]==$pf_finalstate;
	    }
	} else { # make sure to add a blank transitin for placeholder
	    $self->settrans($newfrom,[]);
	}
    }
    $self->{"finalstate"}=$finalstate+$pf_finalstate-1;
    $self->{"penultimate_states"}=[sort keys %new_penultimate_states];
#    printf STDERR ("number of states: %d\n",$#{$self->{"states"}});
#    printf STDERR ("finalstate: %d\n",$self->{"finalstate"});

}

# settrans(from,transition-array): change all of the transitions from a node
#                                  check for monotonicity
sub settrans {
    my $self=shift;
    my $from=shift;
    my $tranarray=shift;
    my $states=$self->{"states"};
    my $tran;
    my $max=-1;

    foreach $tran (@{$tranarray}) {
	if ($from>$tran->[0]) {
	    my($pkg,$file,$line)=caller;
	    warn("onward_fsg::addtrans: from state $from > to state $tran->[0], not adding tranarray; called from $pkg::$file, line $line");
	    return;
	}
	$max=$tran->[0] if $max<$tran->[0];
    }
    if ($max>$self->{"finalstate"}) {
	$self->{"finalstate"}=$max;
	undef $self->{"penultimate_states"};
    }
    &fsg::settrans($self,$from,$tranarray);
}

# reverseindex(): create the revstates array by indexing on the to state
#                 rather than the from state
sub reverseindex {
    my $self=shift;
    my $states=$self->{"states"};
    my $revstates=[];
    my $tran;
    my $from;
    my $to;
    my @trans;
    my @t;

    for($from=0;$from<=$#{$states};$from++) {
	@trans=@{$states->[$from]};
	foreach $tran (@trans) {
	    @t=@{$tran};
	    $to=$t[0];
	    $revstates->[$to]=[] if !defined($revstates->[$to]);
	    push(@{$revstates->[$to]},[$from,@t[1..$#t]]);
	}
    }
    $self->{"revstates"}=$revstates;
}

# findfinal_recurse: recursive iteration for findfinal, should not be called
#                    directly
sub findfinal_recurse {
    my $self=shift;
    my $state=shift;
    my $epsmkr=shift;
    my $final=shift;
    my $prob=shift;
    my $tran;
    my @trans;
    
    if (defined($self->{"revstates"}->[$state])) {
	@trans=@{$self->{"revstates"}->[$state]};
	foreach $tran (@trans) {
	    if ($tran->[2] eq $epsmkr) {
		&findfinal_recurse($self,$tran->[0],$epsmkr,$final,$tran->[1]*$prob);
	    }
	}
    }
    $final->[$state]+=$prob;
}

# @a=findfinal(epsmkr): find all states that can transition via epsilon
#                       transitions to the final state
sub findfinal {
    my $self=shift;
    my $epsmkr=shift;
    my @finalmarkers;
    my $finalstate=$self->{"finalstate"};
    my @uniqmarkers=();
    my $f;

    for($f=0;$f<=$finalstate;$f++) {
	$uniqmarkers[$f]=0;
    }
    $self->reverseindex();
    if (defined($epsmkr)) {
	$self->findfinal_recurse($finalstate,$epsmkr,\@uniqmarkers,1);
    } else {
	$uniqmarkers[$finalstate]=1;
    }
    return @uniqmarkers;
}

# rmepsilon(epsmkr): creates a new fsg with no epsilon transitions
#                    equivalent to the original fsg
sub rmepsilon {
    my $self=shift;
    my $epsmkr=shift;
    my $prunelevel=shift;
    my @newtrans;
    my $to;

    $prunelevel=0.0001 if !defined($prunelevel);
#    print STDERR $prunelevel,"\n";
    my $newself=new onward_fsg();
    my @finalstates=$self->findfinal($epsmkr);

    my $from;
#    print STDERR join(" ","finalstates:",@finalstates),"\n";
    for($from=0;$from<=$#{$self->{"states"}};$from++) {
	@newtrans=&reach($self,$from,$epsmkr,\@finalstates,1,$prunelevel);
	if ($prunelevel) {
	    $newself->{"transprunelevel"}=$prunelevel;
#	    print STDERR "from: $from\n";
	    for ($to=0;$to<=$#newtrans;) {
#		print STDERR join(" ",@{$newtrans[$to]}),"\n";
		if ($newtrans[$to]->[1] < $prunelevel) {
		    splice(@newtrans,$to,1);
		} else {
		    $to++;
		}
	    }
	}
	$newself->settrans($from,[@newtrans]);
    }
    return $newself;
}

# reach: subroutine for rmepsilon
# has no discounting
sub reach {
    my $self=shift;
    my $from=shift;
    my $epsmkr=shift;
    my $finalstates=shift;
    my $prob=shift;
    my $prunelevel=shift;
    my @results=();
    my $states=$self->{"states"};
    my $finalstate=$self->{"finalstate"};
    my $nonullwords=$self->{"nonullwords"};
    my @trans;
    my $tran;
    my @t;

    $nonullwords=1 if !defined($nonullwords);

    if ($prob<$prunelevel) {
	return ();
    }
    if (defined($states->[$from])) {
	@trans=@{$states->[$from]};
	foreach $tran (@trans) {
	    if ($tran->[2] eq $epsmkr) {
		if ($tran->[3] eq "" || !$nonullwords) {
		    my ($r,@reachresults);
		    @reachresults=$self->reach($tran->[0],$epsmkr,
					       $finalstates,$tran->[1]*$prob,$prunelevel);
		    foreach $r (@reachresults) {
			if (ref($r->[3]) eq "ARRAY") {
			    if (ref($tran->[3]) eq "ARRAY") {
				$r->[3]->[0]=$tran->[3]->[0].$r->[3]->[0];
				$r->[3]->[1]*=$tran->[3]->[1];
			    } else {
				$r->[3]->[0]=$tran->[3].$r->[3]->[0];
			    }
			} else {
			    if (ref($tran->[3]) eq "ARRAY") {
				$r->[3]=[$tran->[3]->[0].$r->[3],$tran->[3]->[1]];
			    } else {
				$r->[3]=$tran->[3].$r->[3];
			    }
			}
		    }
		    push(@results,@reachresults);
		}
		
	    } else {
		@t=@{$tran};
		push(@results,
		     [$t[0],$t[1]*$prob,@t[2..$#t]]);
		#propagate NULL transition probabilities
		if ($t[0]!=$finalstate && $finalstates->[$t[0]]) {
		    push(@results,
			 [$finalstate,$t[1]*$prob*$finalstates->[$t[0]],
			  @t[2..$#t]]);
		}
	    }
	}
    }
    return @results;
}
# reach_new: subroutine for rmepsilon
# has bug in it... doeesn't correctly pass NULL transition probabilities
sub reach_new {
    my $self=shift;
    my $from=shift;
    my $epsmkr=shift;
    my $finalstates=shift;
    my $prob=shift;
    my @results=();
    my $states=$self->{"states"};
    my $finalstate=$self->{"finalstate"};
    my $prunelevel=$self->{"transprunelevel"};
    my @trans;
    my $tran;
    my @t;
    my $discount;
    my $newprob;

    if (defined($states->[$from])) {
	@trans=@{$states->[$from]};
	foreach $tran (@trans) {
	    if ($tran->[2] eq $epsmkr) {
		my ($r,@reachresult);
		if ($prunelevel>0 && $tran->[1]*$prob>$prunelevel) {
		    @reachresult=$self->reach($tran->[0],$epsmkr,
					      $finalstates,$tran->[1]*$prob);
		    foreach $r (@reachresult) {
			$r->[3]=$tran->[3].$r->[3];
		    }
		    push(@results,@reachresult);
		}
	    } else {
		@t=@{$tran};
		$discount = ($t[0]!=$finalstate ?
			     1 - $finalstates->[$t[0]] :
			     1);
		$discount /= (1 - $finalstates->[$from]);

		if ($t[1]*$prob*$discount > 0) {
		    push(@results,
			 [$t[0],$t[1]*$prob*$discount,@t[2..$#t]]);
		}
		#propagate NULL transition probabilities
		if ($t[0]!=$finalstate && $finalstates->[$t[0]]) {
		    push(@results,
			 [$finalstate,$t[1]*$prob*$finalstates->[$t[0]],
			  @t[2..$#t]]);
		}
	    }
	}
    }
    return @results;
}

# setlabels(label,field,from,to): sets the label field of the transition
sub setlabels {
    my $self=shift;
    my $label=shift;
    my $field=shift;
    my $from=shift;
    my $to=shift;
    my $tran;
    my $states=$self->{"states"};

    if (!defined($field)||$field<2) {
	$field=3;
    }
    if (defined($from)) {
	foreach $tran (@{$states->[$from]}) {
	    if (defined($to)) {
		$tran->[$field]=$label.$tran->[$field] if ($tran->[0]==$to);
	    } else {
		$tran->[$field]=$label.$tran->[$field];
	    }
	}
    } else {
	for($from=0;$from<=$#{$states};$from++) {
	    foreach $tran (@{$states->[$from]}) {
		$tran->[$field]=$label.$tran->[$field];
	    }
	}
    }
}

# scaleprobs(probscale): scales all probs by this number
# scaleprobs(probscale,from): scales all transitions from this node
# scaleprobs(probscale,from,to): scales all transitions between two nodes
sub scaleprobs {
    my $self=shift;
    my $probscale=shift;
    my $from=shift;
    my $to=shift;
    my $tran;
    my $states=$self->{"states"};

    #safety's sake
    if (!defined($probscale)) {
	$probscale=1;
    }

    if (defined($from)) {
	foreach $tran (@{$states->[$from]}) {
	    if (defined($to)) {
		$tran->[1]*=$probscale if ($tran->[0]==$to);
	    } else {
		$tran->[1]*=$probscale;
	    }
	}
    } else {
	for($from=0;$from<=$#{$states};$from++) {
	    foreach $tran (@{$states->[$from]}) {
		$tran->[1]*=$probscale;
	    }
	}
    }
}

# join_and_print_efsg(fh,magic,@fsgs): print a collection of fsgs as one
#                                      concatenated fsg
sub join_and_print_efsg {
    my $fh=shift;
    my $magic=shift;

    my @fsgs=@_;
    my $totalstates=0;
    my $fsg;

    foreach $fsg (@fsgs) {
	next if $fsg->{"finalstate"}<0;
	$fsg->{"printstart"}=$totalstates;
	$totalstates+=$fsg->{"finalstate"};
    }

    if (ref($fh) eq "GLOB") {
	print $fh "$magic\n";
	print $fh "# Generated by $0 on ".`date`;
	print $fh "# number of states\n";
	print $fh $totalstates+1,"\n";
    } elsif (ref($fh) eq "ARRAY") {
	push (@{$fh}, 
	      "$magic\n",
	      "# Generated by $0 on ".`date`,
	      "# number of states\n",
	      sprintf("%d\n",$totalstates+1));
    } else {
	warn("onward_fsg::printefsg: passed non GLOB or ARRAY reference");
    }

    foreach $fsg (@fsgs) {
	next if $fsg->{"finalstate"}<0;
	$fsg->printefsg_body($fh);
    }
}

# join(@fsgs): join a collection of fsgs together and return one fsg
#              that is the concatenation of all the fsgs.

sub join {
    my $new_fsg=shift;
    my @fsgs=@_;
    my $fsg;
    my $totalstates=$new_fsg->{"finalstate"};
    $totalstates=0 if $totalstates<=0;
    my $states;
    my $s;
    my $tran;
    my @tran;
    my @trans;
    my @penultimate_states=();
    my $finalstate;


    foreach $fsg (@fsgs) {
	next if $fsg->{"finalstate"}<0;
	$states=$fsg->{"states"};
	$finalstate=$fsg->{"finalstate"};
	for ($s=0;$s<=$#{$states};$s++) {
	    $new_fsg->settrans($s+$totalstates,[]);
	    @trans=@{$states->[$s]};
	    foreach $tran (@trans) {
		@tran=@{$tran};
		$new_fsg->addtrans($s+$totalstates,$tran[0]+$totalstates,
				   @tran[1..$#tran]);
		if ($tran[0] == $finalstate && $s!=$finalstate) {
		    push(@penultimate_states,$s+$totalstates);
		}
	    }
	}
	$totalstates+=$fsg->{"finalstate"};
    }
    $new_fsg->{"finalstate"}=$totalstates;
    $new_fsg->{"penultimate_states"}=[@penultimate_states];
    return $new_fsg;
}
    
    
# prune_unreachable(): create a new fsg where all unreachable states
#                      are pruned
sub prune_unreachable {
    my $self=shift;
    my $newself=new onward_fsg();

    my @reachedstates=();
    my @transform=();
    my $from;
    my $tran;
    my @newtran;
    my $states=$self->{"states"};
    my $newstatectr=0;

    $self->reverseindex();
    $self->reach_states(\@reachedstates,$self->{"finalstate"});
    for ($from=0;$from<=$self->{"finalstate"};$from++) {
	if ($reachedstates[$from]) {
	    $transform[$from]=$newstatectr;
	    $newstatectr++;
	} else {
	    $transform[$from]=-1;
	}
    }
    for ($from=0;$from<=$self->{"finalstate"};$from++) {
	if ($transform[$from]>=0) {
	    foreach $tran (@{$states->[$from]}) {
		if ($transform[$tran->[0]]>=0) {
		    @newtran=($transform[$from],@{$tran});
		    $newtran[1]=$transform[$tran->[0]];
		    $newself->addtrans(@newtran);
		}
	    }
	}
    }
    return $newself;
}

# reach_states(): subroutine for prune_unreachable()
sub reach_states {
    my $self=shift;
    my $reachedstates=shift;
    my $curstate=shift;
    my $tran;

    $reachedstates->[$curstate]=1;
    foreach $tran (@{$self->{"revstates"}->[$curstate]}) {
	if (!$reachedstates->[$tran->[0]]) {
	    $self->reach_states($reachedstates,$tran->[0]);
	}
    }
}
    
#make require happy
1;
